/*
 * ion/mod_menu/main.h
 *
 * Copyright (c) Tuomo Valkonen 1999-2007.
 *
 * See the included file LICENSE for details.
 */

#ifndef ION_MOD_MENU_MAIN_H
#define ION_MOD_MENU_MAIN_H

#include <ioncore/binding.h>

extern bool mod_menu_init();
extern void mod_menu_deinit();

extern WBindmap *mod_menu_menu_bindmap;


#endif /* ION_MOD_MENU_MAIN_H */
