/*
 * mod_notionflux/notionflux.h
 *
 * Copyright (c) Tuomo Valkonen 2004-2005.
 *
 * This is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 */

#ifndef NOTIONFLUX_H
#define NOTIONFLUX_H

/* This limit is deliberate, not just laziness. */
#define MAX_SERVED 8
#define CHUNK 1024
#define MAX_DATA (1024*4)

#endif
