/*
 * ion/de/fontset.h
 *
 * Copyright (c) Tuomo Valkonen 1999-2007.
 *
 * See the included file LICENSE for details.
 */

#ifndef ION_DE_FONTSET_H
#define ION_DE_FONTSET_H

#include <ioncore/common.h>
#include <ioncore/gr.h>

extern XFontSet de_create_font_set(const char *fontname);

#endif /* ION_DE_FONTSET_H */
