/*
 * libextl/extl.h
 *
 * Copyright (c) Tuomo Valkonen 1999-2005.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 */

#ifndef LIBEXTL_EXTL_H
#define LIBEXTL_EXTL_H

#include "luaextl.h"

/* Keywords for libextl-mkexports script */

#define EXTL_EXPORT
#define EXTL_EXPORT_AS(T, F)
#define EXTL_EXPORT_MEMBER
#define EXTL_CLASS(CLS, PARCLS)
#define EXTL_SAFE
#define EXTL_UNTRACED

#endif /* LIBEXTL_EXTL_H */
