/*
 * notion/ioncore/profiling.h
 *
 * Copyright (c) The Notion Team 2012.
 *
 * See the included file LICENSE for details.
 */

#ifndef NOTION_IONCORE_PROFILING_H
#define NOTION_IONCORE_PROFILING_H

void ioncore_profiling_start();
void ioncore_profiling_stop();

#endif /* NOTION_IONCORE_PROFILING_H */
