/*
 * ion/ioncore/common.h
 *
 * Copyright (c) Tuomo Valkonen 1999-2007.
 *
 * See the included file LICENSE for details.
 */

#ifndef ION_IONCORE_COMMON_H
#define ION_IONCORE_COMMON_H

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <libtu/types.h>
#include <libtu/output.h>
#include <libtu/misc.h>
#include <libtu/dlist.h>
#include <libtu/obj.h>
#include <libtu/locale.h>
#include <libtu/debug.h>

#include <libextl/extl.h>

#include "../config.h"
#include "classes.h"

#endif /* ION_IONCORE_COMMON_H */
